/*
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of this archive for
 *  more details.
 */

#define DRIVER_NAME "NVVGACON"

#define VT_BUF_HAVE_MEMSETW
#define VT_BUF_HAVE_MEMCPYW
#define VT_BUF_HAVE_MEMMOVEW

#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/console.h>
//#include <linux/console_struct.h>
#include <linux/string.h>
#include <linux/kd.h>
#include <linux/slab.h>
#include <linux/vt_kern.h>
#include <linux/selection.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/io.h>

#define DBG2(s...) if(debug>=2)printk( KERN_DEBUG DRIVER_NAME ": " s)

#define BLANK 0x2007

static spinlock_t vga_lock = SPIN_LOCK_UNLOCKED;

/*
 *  Interface used by the world
 */

static const char *nvvgacon_startup(void);
static void nvvgacon_init(struct vc_data *c, int init);
static void nvvgacon_deinit(struct vc_data *c);
static void nvvgacon_cursor(struct vc_data *c, int mode);
static int nvvgacon_switch(struct vc_data *c);
static int nvvgacon_blank(struct vc_data *c, int blank, int mode_switch);
static int nvvgacon_set_palette(struct vc_data *c, unsigned char *table);
static int nvvgacon_scrolldelta(struct vc_data *c, int lines);
static int nvvgacon_set_origin(struct vc_data *c);
static int nvvgacon_scroll(struct vc_data *c, int t, int b, int dir, int lines);
static u8 nvvgacon_build_attr(struct vc_data *c, u8 color, u8 intensity, u8 blink, u8 underline, u8 reverse);
static void nvvgacon_invert_region(struct vc_data *c, u16 *p, int count);
static void nvvgacon_bmove(struct vc_data *c, int sy, int sx, 
			 int dy, int dx, int height, int width);
static void nvvgacon_clear(struct vc_data *c, int y, int x, 
			  int height, int width);

//static unsigned long nvvgacon_uni_pagedir[2];

struct consw nvvga_con;

/* Description of the hardware situation */
static unsigned long bar0;
static unsigned long bar1;
static char *nvvga_vram_base;		/*  */
static char *video_buf;		/*  */
static char *nvvga_mmio_base0;		/* misc, seq and graphics*/
static char *nvvga_mmio_base1;		/* crt and attribute */
static char *nvvga_mmio_base2;		/* dac */

static int 	video_scan_lines;
static int 	video_font_height;
static unsigned int    nvvga_video_num_columns;	/* Number of text columns */
static unsigned int    nvvga_video_num_lines;	/* Number of text lines */
static unsigned int    nvvga_default_font_height;	/* Height of default screen font */
static unsigned char   nvvga_font_is_default = 1;
static int	       nvvga_vesa_blanked;
static int	       nvvga_palette_blanked;
static int	       nvvga_is_gfx;
static int	       nvvga_video_font_height;
static unsigned int    nvvga_rolled_over = 0;
static char *nvvga_video_port_reg, *seq_port_reg;
static char *nvvga_video_port_val, *seq_port_val;
static char *video_misc_wr, *video_misc_rd;
static char *attrib_port;
static char *dac_reg, *dac_val;

static int debug=0;
static int first_vc = 13;
static int last_vc  = 16;

static struct vc_data	*nvvga_display_fg = NULL;

#ifdef MODULE_PARM
MODULE_PARM(first_vc, "i");
MODULE_PARM(last_vc,  "i");
MODULE_PARM(debug, "i");
#endif

static inline void write_vga(unsigned char reg, unsigned int val)
{
	unsigned int v1, v2;

	v1 = reg + (val & 0xff00);
	v2 = reg + 1 + ((val << 8) & 0xff00);
	writew(v1, nvvga_video_port_reg);
        writew(v2, nvvga_video_port_reg);
}

#undef scr_memsetw
#undef scr_memcpyw
#undef scr_memmovew

static inline void scr_memsetw(u16 *dest, u16 val, unsigned int length) {
    for(length>>=1;length>0;length--) {
        scr_writew(val,dest);
        dest+=4;
    }
}

static inline void scr_memcpyw(u16 *dest, u16 *src, unsigned int length) {
    for(length>>=1;length>0;length--) {
        scr_writew(scr_readw(src),dest);
        dest+=4;
        src+=4;
    }
}

static inline void scr_memmovew(u16 *dest, u16 *src, unsigned int length) {
    if(dest<src)scr_memcpyw(dest,src,length); 
    else {
        length >>= 1;
        dest+=4*length;
        src+=4*length;
        for(;length>0;length--) {
            scr_writew(scr_readw(src),dest);
            dest-=4;
            src-=4;
        }
    }
}

#ifdef MODULE
static const char *nvvgacon_startup(void)
#else
__initfunc(static const char *nvvgacon_startup(void))
#endif
{
    int i;
        
	nvvga_video_num_lines = 25;
	nvvga_video_num_columns = 80;

    video_buf=kmalloc(65536, GFP_KERNEL);

    nvvga_vram_base=ioremap(bar1,8*64*1024);
    nvvga_mmio_base0=ioremap(bar0+0xc0000,1024);
    nvvga_mmio_base1=ioremap(bar0+0x601000,1024);
    nvvga_mmio_base2=ioremap(bar0+0x681000,1024);

    nvvga_video_port_reg=nvvga_mmio_base1+0x3d4;
    nvvga_video_port_val=nvvga_mmio_base1+0x3d5;
    seq_port_reg=nvvga_mmio_base0+0x3c4;
	seq_port_val=nvvga_mmio_base0+0x3c5;
	video_misc_wr=nvvga_mmio_base0+0x3c2;
	video_misc_rd=nvvga_mmio_base0+0x3cc;
	dac_reg=nvvga_mmio_base2+0x3c8;
	dac_val=nvvga_mmio_base2+0x3c9;
	attrib_port=nvvga_mmio_base1+0x3c0;

	for (i=0; i<16; i++) {
		readb(nvvga_mmio_base1+0x3da);
		writeb(i, attrib_port);
		writeb(i, attrib_port);
	}
	writeb(0x20, attrib_port);

	for (i=0; i<16; i++) {
		writeb (color_table[i], dac_reg) ;
		writeb (default_red[i], dac_val) ;
		writeb (default_grn[i], dac_val) ;
		writeb (default_blu[i], dac_val) ;
	}

	nvvga_default_font_height = 16;
	nvvga_video_font_height = 16;
    video_scan_lines = nvvga_video_font_height * nvvga_video_num_lines;
	video_font_height = nvvga_video_font_height;

	return "NVVGACON";
}

static void nvvgacon_init(struct vc_data *c, int init)
{
	
	c->vc_can_do_color = 1;
	c->vc_complement_mask = 0x7700;
	c->vc_display_fg = &nvvga_display_fg;

	if (init) {
		c->vc_cols = nvvga_video_num_columns;
		c->vc_rows = nvvga_video_num_lines;
	} else {
		vc_resize(c->vc_num, nvvga_video_num_columns, nvvga_video_num_lines);
        }
	
	if (nvvga_display_fg == NULL)
		nvvga_display_fg = c;

#if 0
	unsigned long p;
	p = *c->vc_uni_pagedir_loc;
	if (c->vc_uni_pagedir_loc == &c->vc_uni_pagedir ||
	    !--c->vc_uni_pagedir_loc[1])
		con_free_unimap(c->vc_num);
	c->vc_uni_pagedir_loc = nvvgacon_uni_pagedir;
	nvvgacon_uni_pagedir[1]++;
	if (!nvvgacon_uni_pagedir[0] && p)
		con_set_default_unimap(c->vc_num);
#endif
}

static inline void nvvga_set_mem_top(struct vc_data *c)
{
	write_vga(12, (c->vc_visible_origin-c->vc_origin)/2);
}

static void nvvgacon_deinit(struct vc_data *c)
{
#if 0
	if (!--nvvgacon_uni_pagedir[1]) {
		c->vc_visible_origin = nvvga_vram_base;
		nvvga_set_mem_top(c);
		con_free_unimap(c->vc_num);
	}
	c->vc_uni_pagedir_loc = &c->vc_uni_pagedir;
	con_set_default_unimap(c->vc_num);
#endif
	if (nvvga_display_fg == c)
		nvvga_display_fg = NULL;

}

static u8 nvvgacon_build_attr(struct vc_data *c, u8 color, u8 intensity, u8 blink, u8 underline, u8 reverse)
{
	u8 attr = color;

	if (underline)
		attr = (attr & 0xf0) | c->vc_ulcolor;
	else if (intensity == 0)
		attr = (attr & 0xf0) | c->vc_halfcolor;
	if (reverse)
		attr = ((attr) & 0x88) | ((((attr) >> 4) | ((attr) << 4)) & 0x77);
	if (blink)
		attr ^= 0x80;
	if (intensity == 2)
		attr ^= 0x08;
	return attr;
}

static void nvvgacon_invert_region(struct vc_data *c, u16 *p, int count)
{
	while (count--) {
		u16 a = scr_readw(p);
			a = ((a) & 0x88ff) | (((a) & 0x7000) >> 4) | (((a) & 0x0700) << 4);
		scr_writew(a, p);
                p+=4;
	}
}

static void nvvgacon_set_cursor_size(int xpos, int from, int to)
{
	unsigned long flags;
	int curs, cure;
	static int lastfrom, lastto;

	if ((from == lastfrom) && (to == lastto)) return;
	lastfrom = from; lastto = to;

	spin_lock_irqsave(&vga_lock, flags);
	writeb(0x0a, nvvga_video_port_reg);		/* Cursor start */
	curs = readb(nvvga_video_port_val);
	writeb(0x0b, nvvga_video_port_reg);		/* Cursor end */
	cure = readb(nvvga_video_port_val);

	curs = (curs & 0xc0) | from;
	cure = (cure & 0xe0) | to;

        writeb(0x0a, nvvga_video_port_reg);		/* Cursor start */
        writeb(curs, nvvga_video_port_val);
        writeb(0x0b, nvvga_video_port_reg);		/* Cursor end */
        writeb(cure, nvvga_video_port_val);
	spin_unlock_irqrestore( &vga_lock, flags);
}

static void nvvgacon_cursor(struct vc_data *c, int mode)
{
    if (c->vc_origin != c->vc_visible_origin)
	nvvgacon_scrolldelta(c, 0);
    switch (mode) {
	case CM_ERASE:
	    write_vga(14, (64*1024 - 1));
	    break;

	case CM_MOVE:
	case CM_DRAW:
	    write_vga(14, c->vc_x+nvvga_video_num_columns*c->vc_y);
	    switch (c->vc_cursor_type & 0x0f) {
		case CUR_UNDERLINE:
			nvvgacon_set_cursor_size(c->vc_x, 
					video_font_height - (video_font_height < 10 ? 2 : 3),
					video_font_height - (video_font_height < 10 ? 1 : 2));
			break;
		case CUR_TWO_THIRDS:
			nvvgacon_set_cursor_size(c->vc_x, 
					 video_font_height / 3,
					 video_font_height - (video_font_height < 10 ? 1 : 2));
			break;
		case CUR_LOWER_THIRD:
			nvvgacon_set_cursor_size(c->vc_x, 
					 (video_font_height*2) / 3,
					 video_font_height - (video_font_height < 10 ? 1 : 2));
			break;
		case CUR_LOWER_HALF:
			nvvgacon_set_cursor_size(c->vc_x, 
					 video_font_height / 2,
					 video_font_height - (video_font_height < 10 ? 1 : 2));
			break;
		case CUR_NONE:
			nvvgacon_set_cursor_size(c->vc_x, 31, 30);
			break;
          	default:
			nvvgacon_set_cursor_size(c->vc_x, 1, video_font_height);
			break;
		}
	    break;
    }
}

static int nvvgacon_switch(struct vc_data *c)
{
	/*
	 * We need to save screen size here as it's the only way
	 * we can spot the screen has been resized and we need to
	 * set size of freshly allocated screens ourselves.
	 */
	nvvga_video_num_columns = c->vc_cols;
	nvvga_video_num_lines = c->vc_rows;
        return 1;
}

static void nvvga_set_palette(struct vc_data *c, unsigned char *table)
{
	int i, j ;

	for (i=j=0; i<16; i++) {
		writeb (table[i], dac_reg) ;
                writeb (c->vc_palette[j++]>>2, dac_val) ;
                writeb (c->vc_palette[j++]>>2, dac_val) ;
                writeb (c->vc_palette[j++]>>2, dac_val) ;
	}
}

static int nvvgacon_set_palette(struct vc_data *c, unsigned char *table)
{
	if (nvvga_palette_blanked || !CON_IS_VISIBLE(c))
		return -EINVAL;
	nvvga_set_palette(c, table);
	return 0;
}

/* structure holding original VGA register settings */
static struct {
	unsigned char	SeqCtrlIndex;		/* Sequencer Index reg.   */
	unsigned char	CrtCtrlIndex;		/* CRT-Contr. Index reg.  */
	unsigned char	CrtMiscIO;		/* Miscellaneous register */
	unsigned char	HorizontalTotal;	/* CRT-Controller:00h */
	unsigned char	HorizDisplayEnd;	/* CRT-Controller:01h */
	unsigned char	StartHorizRetrace;	/* CRT-Controller:04h */
	unsigned char	EndHorizRetrace;	/* CRT-Controller:05h */
	unsigned char	Overflow;		/* CRT-Controller:07h */
	unsigned char	StartVertRetrace;	/* CRT-Controller:10h */
	unsigned char	EndVertRetrace;		/* CRT-Controller:11h */
	unsigned char	ModeControl;		/* CRT-Controller:17h */
	unsigned char	ClockingMode;		/* Seq-Controller:01h */
} nvvga_state;

static void nvvga_vesa_blank(int mode)
{
	/* save original values of VGA controller registers */
	if(!nvvga_vesa_blanked) {
		spin_lock_irq(&vga_lock);
		nvvga_state.SeqCtrlIndex = readb(seq_port_reg);
		nvvga_state.CrtCtrlIndex = readb(nvvga_video_port_reg);
		nvvga_state.CrtMiscIO = readb(video_misc_rd);
		spin_unlock_irq(&vga_lock);

		writeb(0x00,nvvga_video_port_reg);	/* HorizontalTotal */
		nvvga_state.HorizontalTotal = readb(nvvga_video_port_val);
		writeb(0x01,nvvga_video_port_reg);	/* HorizDisplayEnd */
		nvvga_state.HorizDisplayEnd = readb(nvvga_video_port_val);
		writeb(0x04,nvvga_video_port_reg);	/* StartHorizRetrace */
		nvvga_state.StartHorizRetrace = readb(nvvga_video_port_val);
		writeb(0x05,nvvga_video_port_reg);	/* EndHorizRetrace */
		nvvga_state.EndHorizRetrace = readb(nvvga_video_port_val);
		writeb(0x07,nvvga_video_port_reg);	/* Overflow */
		nvvga_state.Overflow = readb(nvvga_video_port_val);
		writeb(0x10,nvvga_video_port_reg);	/* StartVertRetrace */
		nvvga_state.StartVertRetrace = readb(nvvga_video_port_val);
		writeb(0x11,nvvga_video_port_reg);	/* EndVertRetrace */
		nvvga_state.EndVertRetrace = readb(nvvga_video_port_val);
		writeb(0x17,nvvga_video_port_reg);	/* ModeControl */
		nvvga_state.ModeControl = readb(nvvga_video_port_val);
		writeb(0x01,seq_port_reg);		/* ClockingMode */
		nvvga_state.ClockingMode = readb(seq_port_val);
	}

	/* assure that video is enabled */
	/* "0x20" is VIDEO_ENABLE_bit in register 01 of sequencer */
	spin_lock_irq(&vga_lock);
	writeb(0x01,seq_port_reg);
	writeb(nvvga_state.ClockingMode | 0x20,seq_port_val);

	/* test for vertical retrace in process.... */
	if ((nvvga_state.CrtMiscIO & 0x80) == 0x80)
		writeb(nvvga_state.CrtMiscIO & 0xef,video_misc_wr);

	/*
	 * Set <End of vertical retrace> to minimum (0) and
	 * <Start of vertical Retrace> to maximum (incl. overflow)
	 * Result: turn off vertical sync (VSync) pulse.
	 */
	if (mode & VESA_VSYNC_SUSPEND) {
		writeb(0x10,nvvga_video_port_reg);	/* StartVertRetrace */
		writeb(0xff,nvvga_video_port_val); 	/* maximum value */
		writeb(0x11,nvvga_video_port_reg);	/* EndVertRetrace */
		writeb(0x40,nvvga_video_port_val);	/* minimum (bits 0..3)  */
		writeb(0x07,nvvga_video_port_reg);	/* Overflow */
		writeb(nvvga_state.Overflow | 0x84,nvvga_video_port_val); /* bits 9,10 of vert. retrace */
	}

	if (mode & VESA_HSYNC_SUSPEND) {
		/*
		 * Set <End of horizontal retrace> to minimum (0) and
		 *  <Start of horizontal Retrace> to maximum
		 * Result: turn off horizontal sync (HSync) pulse.
		 */
		writeb(0x04,nvvga_video_port_reg);	/* StartHorizRetrace */
		writeb(0xff,nvvga_video_port_val);	/* maximum */
		writeb(0x05,nvvga_video_port_reg);	/* EndHorizRetrace */
		writeb(0x00,nvvga_video_port_val);	/* minimum (0) */
	}

	/* restore both index registers */
	writeb(nvvga_state.SeqCtrlIndex,seq_port_reg);
	writeb(nvvga_state.CrtCtrlIndex,nvvga_video_port_reg);
	spin_unlock_irq(&vga_lock);
}

static void nvvga_vesa_unblank(void)
{
	/* restore original values of VGA controller registers */
	spin_lock_irq(&vga_lock);
	writeb(nvvga_state.CrtMiscIO,video_misc_wr);

	writeb(0x00,nvvga_video_port_reg);		/* HorizontalTotal */
	writeb(nvvga_state.HorizontalTotal,nvvga_video_port_val);
	writeb(0x01,nvvga_video_port_reg);		/* HorizDisplayEnd */
	writeb(nvvga_state.HorizDisplayEnd,nvvga_video_port_val);
	writeb(0x04,nvvga_video_port_reg);		/* StartHorizRetrace */
	writeb(nvvga_state.StartHorizRetrace,nvvga_video_port_val);
	writeb(0x05,nvvga_video_port_reg);		/* EndHorizRetrace */
	writeb(nvvga_state.EndHorizRetrace,nvvga_video_port_val);
	writeb(0x07,nvvga_video_port_reg);		/* Overflow */
	writeb(nvvga_state.Overflow,nvvga_video_port_val);
	writeb(0x10,nvvga_video_port_reg);		/* StartVertRetrace */
	writeb(nvvga_state.StartVertRetrace,nvvga_video_port_val);
	writeb(0x11,nvvga_video_port_reg);		/* EndVertRetrace */
	writeb(nvvga_state.EndVertRetrace,nvvga_video_port_val);
	writeb(0x17,nvvga_video_port_reg);		/* ModeControl */
	writeb(nvvga_state.ModeControl,nvvga_video_port_val);
	writeb(0x01,seq_port_reg);		/* ClockingMode */
	writeb(nvvga_state.ClockingMode,seq_port_val);

	/* restore index/control registers */
	writeb(nvvga_state.SeqCtrlIndex,seq_port_reg);
	writeb(nvvga_state.CrtCtrlIndex,nvvga_video_port_reg);
	spin_unlock_irq(&vga_lock);
}

static void nvvga_pal_blank(void)
{
	int i;

	for (i=0; i<16; i++) {
		writeb (i, dac_reg) ;
		writeb (0, dac_val) ;
		writeb (0, dac_val) ;
		writeb (0, dac_val) ;
	}
}

static int nvvgacon_blank(struct vc_data *c, int blank, int mode_switch)
{
	switch (blank) {
	case 0:				/* Unblank */
		if (nvvga_vesa_blanked) {
			nvvga_vesa_unblank();
			nvvga_vesa_blanked = 0;
		}
		if (nvvga_palette_blanked) {
			nvvga_set_palette(c, color_table);
			nvvga_palette_blanked = 0;
			return 0;
		}
		nvvga_is_gfx = 0;
		/* Tell console.c that it has to restore the screen itself */
		return 1;
	case 1:				/* Normal blanking */
            	nvvga_pal_blank();
                nvvga_palette_blanked = 1;
		nvvgacon_set_origin(c);
		scr_memsetw((void *)nvvga_vram_base, BLANK, c->vc_screenbuf_size);
		return 1;
	case -1:			/* Entering graphic mode */
		scr_memsetw((void *)nvvga_vram_base, BLANK, c->vc_screenbuf_size);
		nvvga_is_gfx = 1;
		return 1;
	default:			/* VESA blanking */
            	nvvga_vesa_blank(blank-1);
                nvvga_vesa_blanked = blank;
		return 0;
	}
}

/*
 * PIO_FONT support.
 *
 * The font loading code goes back to the codepage package by
 * Joel Hoffman (joel@wam.umd.edu). (He reports that the original
 * reference is: "From: p. 307 of _Programmer's Guide to PC & PS/2
 * Video Systems_ by Richard Wilton. 1987.  Microsoft Press".)
 *
 * Change for certain monochrome monitors by Yury Shevchuck
 * (sizif@botik.yaroslavl.su).
 */

#define cmapsz 8192

static int
nvvgacon_do_font_op(char *arg, int set, int ch512)
{
	int i;
	char *charmap;
	int beg;
//	char * video_port_status = nvvga_video_port_reg + 6;
	int font_select = 0x00;

        charmap = nvvga_vram_base+3;
        beg = 0x06;
	
#ifdef BROKEN_GRAPHICS_PROGRAMS
	/*
	 * All fonts are loaded in slot 0 (0:1 for 512 ch)
	 */

	if (!arg)
		return -EINVAL;		/* Return to default font not supported */

	nvvga_font_is_default = 0;
	font_select = ch512 ? 0x04 : 0x00;
#else	
	/*
	 * The default font is kept in slot 0 and is never touched.
	 * A custom font is loaded in slot 2 (256 ch) or 2:3 (512 ch)
	 */

	if (set) {
		nvvga_font_is_default = !arg;
		if (!arg)
			ch512 = 0;		/* Default font is always 256 */
		font_select = arg ? (ch512 ? 0x0e : 0x0a) : 0x00;
	}

	if ( !nvvga_font_is_default )
		charmap += 4*cmapsz * 8;
#endif

	if (arg) {
		if (set)
			for (i=0; i<cmapsz ; i++)
				writeb(arg[i], charmap + i * 8);
		else
			for (i=0; i<cmapsz ; i++)
				arg[i] = readb(charmap + i * 8);

		/*
		 * In 512-character mode, the character map is not contiguous if
		 * we want to remain EGA compatible -- which we do
		 */

		if (ch512) {
			charmap += 2*cmapsz * 8;
			arg += cmapsz;
			if (set)
				for (i=0; i<cmapsz ; i++)
					writeb(arg[i], charmap + i * 8);
			else
				for (i=0; i<cmapsz ; i++)
					arg[i] = readb(charmap + i * 8);
		}
	}
	
	spin_lock_irq(&vga_lock);
	writeb( 0x00, seq_port_reg );   /* First, the sequencer */
	writeb( 0x01, seq_port_val );   /* Synchronous reset */
	if (set) {
		writeb( 0x03, seq_port_reg ); /* Character Map Select */
		writeb( font_select, seq_port_val );
	}
	writeb( 0x00, seq_port_reg );
	writeb( 0x03, seq_port_val );   /* clear synchronous reset */

	spin_unlock_irq(&vga_lock);

	return 0;
}

/*
 * Adjust the screen to fit a font of a certain height
 */
static int
nvvgacon_adjust_height(unsigned fontheight)
{
	int rows, maxscan;
	unsigned char ovr, vde, fsr;

	if (fontheight == nvvga_video_font_height)
		return 0;

	nvvga_video_font_height = video_font_height = fontheight;

	rows = video_scan_lines/fontheight;	/* Number of video rows we end up with */
	maxscan = rows*fontheight - 1;		/* Scan lines to actually display-1 */

	/* Reprogram the CRTC for the new font size */

	spin_lock_irq(&vga_lock);
	writeb( 0x07, nvvga_video_port_reg );		/* CRTC overflow register */
	ovr = readb(nvvga_video_port_val);
	writeb( 0x09, nvvga_video_port_reg );		/* Font size register */
	fsr = readb(nvvga_video_port_val);
	spin_unlock_irq(&vga_lock);

	vde = maxscan & 0xff;			/* Vertical display end reg */
	ovr = (ovr & 0xbd) +			/* Overflow register */
	      ((maxscan & 0x100) >> 7) +
	      ((maxscan & 0x200) >> 3);
	fsr = (fsr & 0xe0) + (fontheight-1);    /*  Font size register */

	spin_lock_irq(&vga_lock);
	writeb( 0x07, nvvga_video_port_reg );		/* CRTC overflow register */
	writeb( ovr, nvvga_video_port_val );
	writeb( 0x09, nvvga_video_port_reg );		/* Font size */
	writeb( fsr, nvvga_video_port_val );
	writeb( 0x12, nvvga_video_port_reg );		/* Vertical display limit */
	writeb( vde, nvvga_video_port_val );
	spin_unlock_irq(&vga_lock);

//	vc_resize_all(rows, 0);			/* Adjust console size */
	return 0;
}

static int nvvgacon_font_set(struct vc_data *c, struct console_font *font, unsigned flags)
{
	unsigned int charcount = font->charcount;
	int rc;
		
	if (font->width != 8 || (charcount != 256 && charcount != 512))
		return -EINVAL;
	
	rc = nvvgacon_do_font_op(font->data, 1, charcount == 512);
	if (rc)
		return rc;
	
	if (!(flags & KD_FONT_FLAG_DONT_RECALC))
		rc = nvvgacon_adjust_height(font->height);
	return rc;
}	
	

static int nvvgacon_set_origin(struct vc_data *c)
{
	if (nvvga_is_gfx 	/* We don't play origin tricks in graphic modes */
#if 0
	    || (console_blanked && !nvvga_palette_blanked)
#endif
            )	/* Nor we write to blanked screens */
		return 0;
	c->vc_origin = c->vc_visible_origin = (int)video_buf;
	nvvga_set_mem_top(c);
	nvvga_rolled_over = 0;
	return 1;
}

static int nvvgacon_scrolldelta(struct vc_data *c, int lines)
{
    return 0;
}

static int nvvgacon_scroll(struct vc_data *c, int t, int b, int dir, int lines)
{
	if (!lines)
		return 0;
	
	DBG2( "_scroll %i %i %i %i\n", t, b, dir, lines);
	
	if (lines > c->vc_rows)   /* maximum realistic size */
		lines = c->vc_rows;

	switch (dir) {

	case SM_UP:
//	        nvvgacon_bmove(c,lines,0,0,0,c->vc_rows-lines,nvvga_video_num_columns);
	        nvvgacon_bmove(c,t+lines,0,t,0,b-t-lines,nvvga_video_num_columns);
            nvvgacon_clear(c,b-lines,0,lines,nvvga_video_num_columns);
		break;

	case SM_DOWN:
//	        nvvgacon_bmove(c,0,0,lines,0,c->vc_rows-lines,nvvga_video_num_columns);
	        nvvgacon_bmove(c,t,0,t+lines,0,b-t-lines,nvvga_video_num_columns);
            nvvgacon_clear(c,t,0,lines,nvvga_video_num_columns);
		break;
	}

	return 0;
}
/*
 *  The console `switch' structure for the VGA based console
 */

#define NVVGA_ADDR(x,y)  ((u16 *) nvvga_vram_base + (((y)*nvvga_video_num_columns + (x))*4))

static void nvvgacon_putc(struct vc_data *c, int ch, int y, int x)
{
	scr_writew(ch, NVVGA_ADDR(x, y));
}

static void nvvgacon_putcs(struct vc_data *c, const unsigned short *s,
		         int count, int y, int x)
{
	u16 *dest = NVVGA_ADDR(x, y);
	DBG2( "_putcs %i %i %i\n", count, y, x);

	for (; count > 0; count--) {
		scr_writew(*(s++), dest);
        dest+=4;
	}
}

static void nvvgacon_clear(struct vc_data *c, int y, int x, 
			  int height, int width)
{
	u16 *dest = NVVGA_ADDR(x, y);
	u16 eattr = c->vc_video_erase_char;
	DBG2( "_clear %i %i %i %i\n", y, x, height, width);

	if (width <= 0 || height <= 0)
		return;

        for (; height > 0; height--, dest+=nvvga_video_num_columns*4) {
#if 1
            scr_memsetw(dest,eattr,width*2);
#else
            int i;
            for(i=0; i<width; i++)
                scr_writew(eattr, dest+i*4);
#endif
        }
}

static void nvvgacon_bmove(struct vc_data *c, int sy, int sx, 
			 int dy, int dx, int height, int width)
{
	u16 *src, *dest;
    int i;
	DBG2( "_bmove %i %i %i %i %i %i\n", sy, sx, dy, dx, height, width);

	if (width <= 0 || height <= 0) {
		return;
		
	} else if (dy < sy || (dy == sy && dx < sx)) {
		src  = NVVGA_ADDR(sx, sy);
		dest = NVVGA_ADDR(dx, dy);

		for (; height > 0; height--) {
			for(i=0;i<width;i++)scr_writew(scr_readw(src+4*i),dest+4*i);
			src  += nvvga_video_num_columns*4;
			dest += nvvga_video_num_columns*4;
		}
	} else {
		src  = NVVGA_ADDR(sx, sy+height-1);
		dest = NVVGA_ADDR(dx, dy+height-1);

		for (; height > 0; height--) {
			for(i=width-1;i>=0;i--)scr_writew(scr_readw(src+4*i),dest+4*i);
			src  -= nvvga_video_num_columns*4;
			dest -= nvvga_video_num_columns*4;

		}
	}
}

static u16 *nvvgacon_screen_pos( struct vc_data *c, int offset) {
    return (u16 *)c->vc_origin+offset;
}

struct consw nvvga_con = {
	.con_startup = nvvgacon_startup,
	.con_init = nvvgacon_init,
	.con_deinit = nvvgacon_deinit,
	.con_clear = nvvgacon_clear,			/* con_clear */
	.con_putc = nvvgacon_putc,			/* con_putc */
	.con_putcs = nvvgacon_putcs,			/* con_putcs */
	.con_cursor = nvvgacon_cursor,
	.con_scroll = nvvgacon_scroll,		/* con_scroll */
	.con_bmove = nvvgacon_bmove,			/* con_bmove */
	.con_switch = nvvgacon_switch,
	.con_blank = nvvgacon_blank,
//	.con_font_op = nvvgacon_font_op,
	.con_font_set = nvvgacon_font_set,
	.con_set_palette = nvvgacon_set_palette,
	.con_scrolldelta = nvvgacon_scrolldelta,
#ifdef FIX_SCROLL
	.con_set_origin = nvvgacon_set_origin,
	.con_save_screen = nvvgacon_save_screen,
#else
	.con_set_origin =         NULL,
	.con_save_screen =        NULL,
#endif
	.con_build_attr = nvvgacon_build_attr,
	.con_invert_region = nvvgacon_invert_region,
	.con_screen_pos = nvvgacon_screen_pos,
};

#ifndef MODULE
__initfunc(void nvvgacon_setup(char *str, int *ints))
{

	if (ints[0] < 2)
		return;

	if (ints[1] < 1 || ints[1] > MAX_NR_CONSOLES || 
	    ints[2] < 1 || ints[2] > MAX_NR_CONSOLES)
		return;

	first_vc = ints[1]-1;
	last_vc  = ints[2]-1;
}
#endif

#ifdef MODULE
void nvvga_console_init(void)
#else
__initfunc(void nvvga_console_init(void))
#endif
{
	struct pci_dev *dev;

	video_buf=NULL;
	
	if (first_vc > last_vc)
		return;

	dev=pci_find_device(PCI_VENDOR_ID_NVIDIA_SGS, PCI_ANY_ID, NULL);
	if(dev==NULL) {
		printk(KERN_WARNING "No nv3 card found.\n");
		return;
	}
	bar0=pci_resource_start(dev, 0);
	bar1=pci_resource_start(dev, 1);

	DBG2("Found nv3 card,  iomem at 0x%08lx,  vidmem at 0x%08lx.\n", bar0, bar1);
	
	take_over_console(&nvvga_con, first_vc-1, last_vc-1, 0);
}


#ifdef MODULE

int init_module(void)
{
	nvvga_console_init();

	return 0;
}

void cleanup_module(void)
{
	give_up_console(&nvvga_con);
    
	if(video_buf!=NULL) {
		kfree(video_buf);
		iounmap(nvvga_vram_base);
		iounmap(nvvga_mmio_base0);
		iounmap(nvvga_mmio_base1);
		iounmap(nvvga_mmio_base2);
	}

}

MODULE_LICENSE("GPL");

#endif
